@@ -24,6 +24,8 @@ packet.h - Packet definition for `ttyrpld` and the rpldev module
 #ifndef RPL_PACKET_H
 #define RPL_PACKET_H 1
 
+#define PACKED __attribute__((packed))
+
 #ifdef __KERNEL__
 #    include <linux/types.h>
 #else
@@ -42,17 +44,27 @@ enum {
     EV_WRITE,
     EV_IOCTL,
     EV_CLOSE,
+    EV_IDENT = 240,
+    EV_IGNORE = 255,
 };
 
 struct rpld_packet {
     uint32_t dev;
-    uint16_t size;
-    uint8_t event, magic;
+    uint16_t size PACKED;
+    uint8_t event PACKED, magic PACKED;
 };
 
-struct log_packet {
-    struct rpld_packet p;
-    struct timeval tv __attribute__((packed));
+struct mem_packet {
+    uint32_t dev;
+    uint16_t size PACKED;
+    uint8_t event PACKED, magic PACKED;
+    struct timeval tv PACKED;
+};
+
+struct disk_packet {
+    uint16_t size;
+    uint8_t event PACKED, magic PACKED;
+    struct timeval tv PACKED;
 };
 
 #endif /* RPL_PACKET_H */
