@@ -24,12 +24,28 @@ shared.h
 #ifndef RPL_USHARED_H
 #define RPL_USHARED_H 1
 
+#include <endian.h>
 #include <stdint.h>
 
+#ifndef O_BINARY
+#    define O_BINARY 0
+#endif
+#define MAXFNLEN 1024
+
+#if __BYTE_ORER == __BIG_ENDIAN
+#    define BIGENDIAN 1
+#endif
+
+extern void SH_check_balign(void);
 extern const char *SH_event_type[];
 extern int SH_count_args(const char **);
-extern char *SH_devname(uint32_t, char *, size_t);
-extern char *SH_devname_fs(uint32_t, char *, size_t);
+extern ssize_t SH_read_wait(int, void *, size_t);
+extern off_t SH_skip(int, off_t, int);
+extern void SH_swab(void *, size_t);
+
+inline static unsigned int min_uint(unsigned int a, unsigned int b) {
+    return a < b ? a : b;
+}
 
 #endif /* RPL_SHARED_H */
 
