@@ -45,9 +45,10 @@ user/replay.c - Realtime TTY log analyzer
 #define BUFSIZE     4096
 
 enum { // current follow mode state:
-    FM_NONE    = 0,
-    FM_CATCHUP = 1, // playing old content
-    FM_LIVE    = 2, // already hit EOF once and waiting for new data
+    FM_NONE = 0,
+    FM_CATCHUP, // playing old content
+    FM_LIVE,    // already hit EOF once and waiting for new data
+    FM_NODELAY, // same as FM_LIVE, but different code parts are executed
 };
 
 enum { // skip types
@@ -58,7 +59,7 @@ enum { // skip types
 
 struct {
     double factor;
-    int follow;
+    int follow, nodelay;
     long ovcorr;
     struct {
         long val;
@@ -67,6 +68,7 @@ struct {
 } Opt = {
     .factor   = 1.0,
     .follow   = 0,
+    .nodelay  = 0,
     .seek_to  = {
         .type = SKIP_NONE,
     },
@@ -77,26 +79,21 @@ static void replay_file(int);
 static unsigned long calc_ovcorr(unsigned long, int) __attribute__((unused));
 static int find_next_packet(int);
 static int get_options(int *, const char ***);
-inline static int imin(int, int);
-static ssize_t read_null(int, size_t);
+inline static ssize_t read_nullfm(int, size_t);
 static ssize_t read_through(int, int, size_t);
-static ssize_t read_wait(int, void *, size_t);
-static ssize_t read_waitfm(int, void *, size_t, int *);
+static ssize_t read_waitfm(int, void *, size_t);
 static int seek_to_end(int);
-static int usleep_std(struct timeval *, long *);
 static int usleep_ovcorr(struct timeval *, long *);
 static void tv_delta(const struct timeval *, const struct timeval *,
  struct timeval *);
 inline static unsigned long long tv2usec(const struct timeval *);
 inline static void usec2tv(unsigned long long, struct timeval *);
 
-static int (*usleep_chosen)(struct timeval *, long *) = usleep_ovcorr;
-
 //-----------------------------------------------------------------------------
 int main(int argc, char **argv) {
     if(!get_options(&argc, (const char ***)&argv)) { return EXIT_FAILURE; }
 
-    printf(
+    fprintf(stderr,
       "> ttyreplay\n"
       "This program comes with ABSOLUTELY NO WARRANTY; it is free software\n"
       "and you are welcome to redistribute it under certain conditions;\n"
@@ -109,9 +106,8 @@ int main(int argc, char **argv) {
     while(*argv != NULL) {
         int fd;
         if((fd = open(*argv, O_RDONLY)) < 0) {
-            fprintf(stderr, "Could not open %s: %s\n",
-             *argv, strerror(errno));
-            ++argv;
+            fprintf(stderr, "Could not open %s: %s\n", *argv++,
+             strerror(errno));
             continue;
         }
 
@@ -124,7 +120,7 @@ int main(int argc, char **argv) {
 }
 
 static void replay_file(int fd) {
-    struct log_packet packet;
+    struct disk_packet packet;
     struct timeval stamp;
     ssize_t eax;
     long skew = 0;
@@ -138,11 +134,11 @@ static void replay_file(int fd) {
         }
     }*/
 
-    while((eax = read_waitfm(fd, &packet, sizeof(struct log_packet),
-     &Opt.follow)) == sizeof(struct log_packet)) {
+    while((eax = read_waitfm(fd, &packet, sizeof(struct disk_packet))) ==
+     sizeof(struct disk_packet)) {
         struct timeval delta;
 
-        if(packet.p.magic != MAGIC_SIG) {
+        if(packet.magic != MAGIC_SIG) {
             fprintf(stderr, "\n" "<Packet inconsistency! "
              "Trying to find next valid packet.>\n");
             tick = 0;
@@ -152,31 +148,36 @@ static void replay_file(int fd) {
             continue;
         }
 
-        if(!tick) {
-            // No delay after the first packet has been read...
-            ++tick;
-        } else if(Opt.follow != FM_LIVE) {
-            /* ... only when we know the next packet and can calculate the
-            time difference. We do not execute this code in FM_LIVE mode,
-            because we are already waiting enough with read_wait(). In fact,
-            allowing this block for FM_LIVE creates a skew! */
-            tv_delta(&stamp, &packet.tv, &delta);
-            if(Opt.factor != 1.0) {
-                usec2tv(tv2usec(&delta) / Opt.factor, &delta);
-            }
-            usleep_chosen(&delta, &skew);
-        }
-
-        memcpy(&stamp, &packet.tv, sizeof(struct timeval));
-        switch(packet.p.event) {
+        switch(packet.event) {
             case EV_WRITE:
-                read_through(fd, STDOUT_FILENO, packet.p.size);
+                if(!tick) {
+                    // No delay after the first packet has been read...
+                    ++tick;
+                } else if(Opt.follow != FM_LIVE && Opt.follow != FM_NODELAY) {
+                    /* ... only when we know the next packet and can calculate
+                    the time difference. We do not execute this code in FM_LIVE
+                    mode, because we are already waiting enough with
+                    SH_read_wait(). In fact, allowing this code block for
+                    FM_LIVE creates a skew! */
+                    tv_delta(&stamp, &packet.tv, &delta);
+                    if(Opt.factor != 1.0) {
+                        usec2tv(tv2usec(&delta) / Opt.factor, &delta);
+                    }
+                    usleep_ovcorr(&delta, &skew);
+                }
+                memcpy(&stamp, &packet.tv, sizeof(struct timeval));
+                read_through(fd, STDOUT_FILENO, packet.size);
+                break;
+            case EV_IDENT:
+                fprintf(stderr, "\n" "Created using ");
+                read_through(fd, STDERR_FILENO, packet.size);
+                fprintf(stderr, "\n");
                 break;
             case EV_CLOSE:
                 fprintf(stderr, "\n" "<tty device has been closed>\n");
                 // fallthrough
             default:
-                read_null(fd, packet.p.size);
+                read_nullfm(fd, packet.size);
                 break;
         }
     }
@@ -212,15 +213,15 @@ static unsigned long calc_ovcorr(unsigned long ad, int rd) {
 }
 
 static int find_next_packet(int fd) {
-#define LZ           sizeof(struct log_packet)
+#define LZ           sizeof(struct disk_packet)
 #define BZ           (2 * LZ)
-#define MAGIC_OFFSET offsetof(struct log_packet, p.magic)
+#define MAGIC_OFFSET offsetof(struct disk_packet, magic)
     char buf[BZ];
-    struct log_packet *packet = (void *)buf;
+    struct disk_packet *packet = (void *)buf;
     size_t s;
     int ok = 0;
 
-    read_wait(fd, buf, BZ);
+    if(read_waitfm(fd, buf, BZ) < BZ) { return 0; }
 
     while(1) {
         char *ptr;
@@ -232,45 +233,50 @@ static int find_next_packet(int fd) {
         if((ptr = memchr(buf, MAGIC_SIG, BZ)) != NULL &&
          ptr - buf >= MAGIC_OFFSET) {
             // A magic byte has been found and the packet start is complete
-            char *ctx = containerof(ptr, struct log_packet, p.magic);
+            char *ctx = containerof(ptr, struct disk_packet, magic);
             if(ctx != buf) {
                 size_t cnt = buf + BZ - ctx;
                 ctx = memmove(buf, ctx, cnt);
-                read_wait(fd, buf + cnt, BZ - cnt);
+                if(read_waitfm(fd, buf + cnt, BZ - cnt) < BZ - cnt) {
+                    return 0;
+                }
             }
         } else if(ptr != NULL) {
             /* Magic byte, but of no use. Discard it, and fill up with new
             data from the descriptor. */
             size_t cnt = buf + BZ - ptr - 1;
             memmove(buf, ptr + 1, cnt);
-            read_wait(fd, buf + cnt, BZ - cnt);
+            if(read_waitfm(fd, buf + cnt, BZ - cnt) < BZ - cnt) { return 0; }
             continue;
         } else {
             /* No magic byte, but since it might be just the next byte in
             the stream, only read LZ bytes. */
             memmove(buf, buf + BZ - LZ, BZ - LZ);
-            read_wait(fd, buf + LZ, BZ - LZ);
+            if(read_waitfm(fd, buf + LZ, BZ - LZ) < BZ - LZ) { return 0; }
             continue;
         }
 
-        s = packet->p.size;
+        s = packet->size;
         if(s > 4096) {
             /* The default tty buffer size is 4096, and any size above this is
             questionable at all. Start with a new slate. */
             ok = 0;
             memmove(buf, buf + LZ, BZ - LZ);
-            read_wait(fd, buf, BZ);
+            if(read_waitfm(fd, buf, BZ) < BZ) { return 0; }
             continue;
         }
 
         if(s < LZ) {
             memmove(buf, buf + LZ + s, BZ - LZ - s);
-            read_wait(fd, buf + BZ - LZ - s, LZ + s);
+            if(read_waitfm(fd, buf + BZ - LZ - s, LZ + s) < LZ + s) {
+                return 0;
+            }
         } else {
-            /* There is no header (according to p.size) in our buffer, so we
+            /* There is no header (according to .size) in our buffer, so we
             can blindly munge lots of data. */
-            read_null(fd, s - LZ);
-            read_wait(fd, buf, BZ);
+            if(!read_nullfm(fd, s - LZ) || read_waitfm(fd, buf, BZ) < BZ) {
+                return 0;
+            }
         }
 
         if((ptr = memchr(buf, MAGIC_SIG, MAGIC_OFFSET + 1)) == NULL ||
@@ -287,15 +293,15 @@ static int find_next_packet(int fd) {
     fprintf(stderr, "\n" "<Found packet boundary>\n");
 
     // Finally adjust the read pointer to a packet boundary
-    if((s = packet->p.size) < 16) {
+    if((s = packet->size) < 16) {
         /* Mmhkay, there is another header other than the current (packet)
         in the buffer. Crap, another one gone. */
         memmove(buf, buf + LZ + s, BZ - LZ - s);
-        read_wait(fd, buf + BZ - LZ - s, s);
-        read_null(fd, packet->p.size);
+        if(read_waitfm(fd, buf + BZ - LZ - s, s) < s) { return 0; }
+        if(!read_nullfm(fd, packet->size)) { return 0; }
     } else {
         // Just subtract what we have already read into the buffer
-        read_null(fd, s - LZ);
+        if(!read_nullfm(fd, s - LZ)) { return 0; }
     }
 
     return 1;
@@ -306,21 +312,21 @@ static int find_next_packet(int fd) {
 
 static int get_options(int *argc, const char ***argv) {
     static const char *_empty_argv[] = {NULL, NULL};
+    char *tmp;
     struct poptOption options_table[] = {
-#define LONGER_OPTION(a, b) (((a) << 8) | (b))
-        {"std", 0, POPT_ARG_NONE, NULL, LONGER_OPTION('-', 's'),
-         "Do not use sleep() overhead correction", NULL},
         {NULL, 'F', POPT_ARG_NONE, NULL, 'F',
          "Live feed follow mode (like `tail -f`)", NULL},
-        {NULL, 'J', POPT_ARG_STRING, &Opt.seek_to, 'J',
+/*        {NULL, 'J', POPT_ARG_STRING, &tmp, 'J',
          "Seek to time relative to the start of the log"
-         " (HH:MM:SS or total SEC)", "tspec"},
+         " (HH:MM:SS or total SEC)", "tspec"},*/
         {NULL, 'S', POPT_ARG_DOUBLE, &Opt.factor, '\0',
          "Speed factor (default: 1.0)", NULL},
+        {NULL, 'T', POPT_ARG_NONE, &Opt.nodelay, 'T',
+         "No delays (useful for text file grepping)", NULL},
         {NULL, 'f', POPT_ARG_NONE, NULL, 'f',
          "Catch-up follow mode (play file, switch to live feed on EOF)", NULL},
-        {NULL, 'j', POPT_ARG_LONG, &Opt.seek_to.val, 'j',
-         "Skip the given number of EV_WRITE packets", NULL},
+/*        {NULL, 'j', POPT_ARG_LONG, &Opt.seek_to.val, 'j',
+         "Skip the given number of EV_WRITE packets", NULL},*/
         POPT_AUTOHELP
         POPT_TABLEEND
     };
@@ -332,16 +338,34 @@ static int get_options(int *argc, const char ***argv) {
     ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
     while((c = poptGetNextOpt(ctx)) >= 0) {
         switch(c) {
-            case LONGER_OPTION('-', 's'):
-                /* Don't use it, it just does not work very nice with high -S
-                speeds and short delays. */
-                usleep_chosen = usleep_std;
-                break;
             case 'F':
                 Opt.follow = FM_LIVE;
                 break;
-            case 'J':
+            case 'J': {
+                char *sec = "", *min = "", *hr = "";
+
                 Opt.seek_to.type = SKIP_TIME;
+                Opt.seek_to.val  = 0;
+
+                if((sec = strrchr(tmp, ':')) == NULL) {
+                    sec = tmp;
+                } else {
+                    *sec++ = '\0';
+                    if((min = strrchr(tmp, ':')) == NULL) {
+                        min = tmp;
+                    } else {
+                        *min++ = '\0';
+                        hr = tmp;
+                    }
+                }
+                Opt.seek_to.val = strtoul(sec, NULL, 0) +
+                 60 * strtoul(min, NULL, 0) + 3600 * strtoul(hr, NULL, 0);
+                free(tmp);
+                break;
+            }
+            case 'T':
+                Opt.follow = FM_NODELAY;
+                break;
             case 'f':
                 Opt.follow = FM_CATCHUP;
                 break;
@@ -367,28 +391,6 @@ static int get_options(int *argc, const char ***argv) {
 
     poptFreeContext(ctx);
     return 1;
-#undef LONGER_OPTION
-}
-
-inline static int imin(int a, int b) {
-    return a < b ? a : b;
-}
-
-static ssize_t read_null(int fd, size_t count) {
-    /* Read and discard data (i.e. do not return it to the caller).
-    Nice workaround for lseek() on unseekable descriptors. */
-    char buf[BUFSIZE];
-    size_t rem = count;
-
-    while(rem > 0) {
-        ssize_t eax = read(fd, buf, imin(BUFSIZE, rem));
-        if(eax < 0) { return 0; }
-        if(eax == rem) { break; }
-        rem -= eax;
-        sched_yield();
-    }
-
-    return count;
 }
 
 static ssize_t read_through(int in, int out, size_t count) {
@@ -397,7 +399,7 @@ static ssize_t read_through(int in, int out, size_t count) {
     size_t rem = count;
 
     while(rem > 0) {
-        ssize_t eax = read(in, buf, imin(BUFSIZE, rem));
+        ssize_t eax = read(in, buf, min_uint(BUFSIZE, rem));
         if(eax < 0) { return 0; }
         write(out, buf, eax);
         if(eax == rem) { break; }
@@ -408,47 +410,34 @@ static ssize_t read_through(int in, int out, size_t count) {
     return count;
 }
 
-static ssize_t read_wait(int fd, void *buf, size_t count) {
-    /* A wrapper for read() which guarantees that all bytes requested will
-    be in BUF after read_wait() returns. (Except if there is an error.)
-    Note that it will retry to read when it hits EOF, so only use this on files
-    which are still being written to! */
-    size_t rem = count;
-
-    while(rem > 0) {
-        ssize_t eax = read(fd, buf, rem);
-        if(eax < 0) { return 0; }
-        if(eax == rem) { break; }
-        buf += eax;
-        rem -= eax;
-        sched_yield(); // usleep(1) could also be a possibility
-    }
-    return count;
+inline static ssize_t read_nullfm(int fd, size_t count) {
+    return SH_skip(fd, count, Opt.follow != FM_NONE && Opt.follow != FM_NODELAY);
 }
 
-static ssize_t read_waitfm(int fd, void *buf, size_t count, int *follow) {
-    /* Wrapper function for either read() or read_wait(). If EOF has been
+static ssize_t read_waitfm(int fd, void *buf, size_t count) {
+    /* Wrapper function for either read() or SH_read_wait(). If EOF has been
     detected (ATM), we switch into FM_LIVE mode (if currently in FM_CATCHUP
     mode). */
     struct stat sb;
     off_t pos;
 
-    if(!*follow) {
+    if(Opt.follow == FM_NONE || Opt.follow == FM_NODELAY) {
         /* If no follow mode is selected, the while() loop in replay_file()
         shall terminate as soon as we encounter EOF or have a short read (i.e.
-        read less then conut). */
+        read less then count). */
         return read(fd, buf, count);
     }
 
     /* If it is a live feed (follow mode), complete the read (i.e. read all
-    requested bytes. */
-    if(*follow != FM_LIVE && fstat(fd, &sb) == 0 &&
-     (pos = lseek(fd, 0, SEEK_CUR)) != 0 && pos == sb.st_size) {
+    requested bytes. fstat() must come after lseek() -- in the very case that
+    the file size increases inbetween. */
+    if(Opt.follow == FM_CATCHUP && (pos = lseek(fd, 0, SEEK_CUR)) != 0 &&
+     fstat(fd, &sb) == 0 && pos == sb.st_size) {
         fprintf(stderr, "\n" "<Switching to live feed follow mode>\n");
-        *follow = FM_LIVE;
+        Opt.follow = FM_LIVE;
     }
 
-    return read_wait(fd, buf, count);
+    return SH_read_wait(fd, buf, count);
 }
 
 static int seek_to_end(int fd) {
@@ -464,7 +453,7 @@ static int seek_to_end(int fd) {
 
         while(read(fd, buf, BUFSIZE) > 0);
         if(errno != EAGAIN) {
-            perror("Error while read()ing");
+            perror("read()");
             return 0;
         }
         fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) & ~O_NONBLOCK);
@@ -473,14 +462,6 @@ static int seek_to_end(int fd) {
     return find_next_packet(fd);
 }
 
-static int usleep_std(struct timeval *req, long *skew) {
-    struct timespec nano_req = {
-        .tv_sec  = req->tv_sec,
-        .tv_nsec = req->tv_usec * 1000,
-    };
-    return nanosleep(&nano_req, NULL);
-}
-
 static int usleep_ovcorr(struct timeval *req, long *skew) {
     /* The manual page for nanosleep(2) says this under the "BUGS" section:
 
