@@ -26,16 +26,19 @@ user/rpld.c - User space daemon (filtering, etc.)
 #include <sys/time.h> // gettimeofday()
 #include <sys/types.h>
 #include <ctype.h>
+#include <endian.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <popt.h>
 #include <pwd.h>
 #include <signal.h>
+#include <stdarg.h>
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 extern int posix_memalign(void **, size_t, size_t);
 #include <string.h>
+#include <syslog.h>
 #include <time.h>
 #include <unistd.h>
 
@@ -44,24 +47,27 @@ extern int posix_memalign(void **, size_t, size_t);
 #include "rpl_ioctl.h"
 #include "rpl_packet.h"
 #include "ushared.h"
-#define MAXFNLEN 1024
+
+enum {
+    NF_NEWLINE = 1 << 0,
+};
 
 struct tty {
     uint32_t dev;
-    size_t in, out;
     int fd;
 };
 
 // The prototype order is the same as the function body order
 static void mainloop(int);
 
-static int packet_preaction(struct log_packet *, struct tty *);
-static int packet_exclude(struct log_packet *);
-static int packet_action(struct log_packet *, struct tty *, int);
+static int packet_preaction(struct mem_packet *);
+static int packet_action(struct mem_packet *, struct tty *, int);
 
+static char *devname(uint32_t, char *, size_t);
+static char *devname_fs(uint32_t, char *, size_t);
 static struct tty *get_tty(uint32_t);
 static int log_open(struct tty *);
-static void log_write(struct log_packet *, struct tty *, int);
+static void log_write(struct mem_packet *, struct tty *, int);
 static void log_close(struct tty *);
 
 static int init_device(const char *);
@@ -74,28 +80,30 @@ static void sighandler_alrm(int);
 
 static int get_options(int *, const char ***);
 static int read_config(const char *);
+static int read_config_bp(const char *, const char *);
 static int pconfig_user(const char *, unsigned char, void *, void *);
 
 static char *getnamefromuid(uid_t, char *, size_t);
 static uid_t getuidfromname(const char *);
-inline static int imin(int, int);
+static int notify(int, unsigned int, const char *, ...);
 static char *replace(char *, const char *, size_t, char **);
 static int value_cmp(void *, void *);
 
 // Global variables (begin with uppercase character)
 static struct {
-    long bsize, fbsize;
-    char *ovdev, *fbdev, *ofmt;
-    int verbose;
-    long user_id;
+    char *_version, *ovdev, *fbdev, *ofmt;
+    int nocreat, syslog, verbose;
+    long bsize, user_id;
 } Opt = {
-    .bsize    = -1,
-    .fbsize   = 16384,
+    ._version = "0.85",
+    .bsize    = 16384,
     .ovdev    = NULL,
     .fbdev    = "/dev/rpl:/dev/misc/rpl",
-    .ofmt     = "%u.%d.%l",
+    .nocreat  = 0,
+    .ofmt     = "%u@%d.%t.%l",
+    .syslog   = 0,
     .verbose  = 0,
-    .user_id = -1,
+    .user_id  = -1,
 };
 
 struct {
@@ -103,31 +111,32 @@ struct {
     unsigned long long in, out;
 } Stats;
 
-static void *Buffer = NULL;
 static struct HXbtree *Ttys = NULL;
+static void *Buffer = NULL;
 static int Running = 1;
 
 //-----------------------------------------------------------------------------
-int main(int argc, char **argv) {
+int main(int argc, const char **argv) {
     int fd;
 
-    if(!read_config("rpld.conf") ||
-     !get_options(&argc, (const char ***)&argv)) {
+    if(!read_config_bp(*argv, "rpld.conf") || !get_options(&argc, &argv)) {
         return EXIT_FAILURE;
     }
 
     if(Opt.verbose) {
         memset(&Stats, 0, sizeof(Stats));
         printf(
-          "# rpld\n"
+          "# rpld %s\n"
           "This program comes with ABSOLUTELY NO WARRANTY; it is free"
           " software\n"
           "and you are welcome to redistribute it under certain conditions;\n"
           "for details see the doc/GPL2.txt file which should have come with\n"
-          "this program.\n\n"
+          "this program.\n\n",
+          Opt._version
         );
     }
 
+    SH_check_balign();
     if((Ttys = HXbtree_init(HXBT_ASSOC | HXBT_LKEY | HXBT_FCMP,
      value_cmp)) == NULL) {
         perror("HXbtree_init()");
@@ -141,12 +150,14 @@ int main(int argc, char **argv) {
 
     if(init_sbuffer(fd) < 0) { return EXIT_FAILURE; }
     init_sighandler();
-    if(Opt.verbose) { alarm(1); } // Callback tick for statistics
 
-    if(Opt.user_id >= 0) {
-        // Not seteuid(), because it shall not be posibble to regain privs.
-        setuid(Opt.user_id);
-    }
+    // Callback tick for statistics
+    if(Opt.verbose) { alarm(1); }
+
+    // Not seteuid(), because it shall not be posibble to regain privs.
+    if(Opt.user_id >= 0) { setuid(Opt.user_id); }
+
+    if(Opt.syslog) { openlog("rpld", LOG_PID, LOG_DAEMON); }
 
     mainloop(fd);
     close(fd);
@@ -155,57 +166,94 @@ int main(int argc, char **argv) {
 }
 
 static void mainloop(int fd) {
+    ssize_t (*reader)(int, void *, size_t) = read;
+    int pk = 0;
+
+    { /* read() on the device might return less than we request, so we use
+      SH_read_wait() and loop until there is data. However, we do not want to
+      wait on e.g. pipes. */
+      struct stat sb;
+      fstat(fd, &sb);
+      if(S_ISCHR(sb.st_mode)) { reader = SH_read_wait; }
+    }
+
     while(Running) {
-        struct log_packet packet;
+        struct mem_packet packet;
         struct tty *tty;
 
-        read(fd, &packet.p, sizeof(struct rpld_packet));
-
-        if(packet.p.magic == MAGIC_2_6) {
+        errno = 0;
+        if(reader(fd, &packet, sizeof(struct rpld_packet)) <
+         sizeof(struct rpld_packet)) {
+            Running = 0;
+            break;
+        }
+++pk;
+        if(packet.magic == MAGIC_2_6) {
             // nothing needed
-        } else if(packet.p.magic == MAGIC_2_4) {
-            packet.p.dev = K26_MKDEV(K24_MAJOR(packet.p.dev),
-             K24_MINOR(packet.p.dev));
-            packet.p.magic = MAGIC_2_6;
+        } else if(packet.magic == MAGIC_2_4) {
+            packet.magic = MAGIC_2_6;
+            packet.dev   = K26_MKDEV(K24_MAJOR(packet.dev),
+             K24_MINOR(packet.dev));
         } else {
             ++Stats.badpack;
-            if(Opt.verbose) {
-                printf("\n" "Bogus packet (magic is 0x%X)!\n",
-                 packet.p.magic);
-            }
+            notify(LOG_WARNING, NF_NEWLINE, "Bogus packet (magic is 0x%X) pack%d!\n",
+             packet.magic, pk);
             continue;
         }
 
         /* Timestamp is taken here instead of in the Kernel module to reduce
         the amount of data which has to be copied over the device. */
         gettimeofday(&packet.tv, NULL);
-        if((tty = get_tty(packet.p.dev)) == NULL ||
-         !packet_preaction(&packet, tty) || packet_exclude(&packet)) {
-            lseek(fd, packet.p.size, SEEK_CUR);
+
+        if(!packet_preaction(&packet) || (tty = get_tty(packet.dev)) == NULL) {
+            SH_skip(fd, packet.size, 0);
             continue;
         }
 
         packet_action(&packet, tty, fd);
     }
 
+    if(Opt.verbose) {
+        if(errno == 0) { fprintf(stderr, "<EOF on input>\n"); }
+        else { perror("read()"); }
+    }
     return;
 }
 
 //-----------------------------------------------------------------------------
-static int packet_preaction(struct log_packet *packet, struct tty *tty) {
-    switch(packet->p.event) {
+static int packet_preaction(struct mem_packet *packet) {
+    /* pty masters are an exact copy of their slaves,
+    so no need to record them too. */
+    register long maj = K26_MAJOR(packet->dev);
+
+    if(maj == 2 /* BSD ptm (/dev/ptypXY) */) {
+        if(packet->event != EV_CLOSE) { return 0; }
+        packet->dev = K26_MKDEV(K26_MAJOR(3), K26_MINOR(packet->dev));
+    }
+
+    if(maj >= 128 && maj < 136 /* Unix98 ptm (-) */) {
+        if(packet->event != EV_CLOSE) { return 0; }
+        /* When closing an xterm window or ssh session, the EV_CLOSE event
+        actually happens on the master side. To keep it easy, it's device
+        signature is simply changed to the slave. */
+        packet->dev = K26_MKDEV(K26_MAJOR(packet->dev) + 8,
+         K26_MINOR(packet->dev));
+    }
+
+    /* The excluded ptm packets above are not counted at all, whereas
+    "ignored" packets are. See below for ignore rules. */
+
+    switch(packet->event) {
         case EV_OPEN:
             ++Stats.open;
             break;
         case EV_READ:
             ++Stats.read;
-            Stats.in += packet->p.size;
-            tty->in  += packet->p.size;
+            Stats.in += packet->size;
             break;
         case EV_WRITE:
             ++Stats.write;
-            Stats.out += packet->p.size;
-            tty->out  += packet->p.size;
+            Stats.out += packet->size;
             break;
         case EV_IOCTL:
             ++Stats.ioctl;
@@ -214,32 +262,22 @@ static int packet_preaction(struct log_packet *packet, struct tty *tty) {
             ++Stats.close;
             break;
         default:
-            if(Opt.verbose) {
-                printf("\n" "Unknown packet type 0x%X\n",
-                 packet->p.event);
-            }
+            notify(LOG_WARNING, NF_NEWLINE, "Unknown packet type 0x%X\n",
+             packet->event);
             return 0;
     }
+
+    if(packet->event == EV_IOCTL) {
+        /* ioctl requests are actually logged by the rpldev Kernel module,
+        but I do not see a use for them ATM. */
+        return 0;
+    }
+
     return 1;
 }
 
-static int packet_exclude(struct log_packet *packet) {
-    register long maj;
-    if(packet->p.event == EV_IOCTL) {
-        /* ioctl requests are actually logged by the rpldev Kernel module,
-        but I do not see a use for them ATM. */
-        return 1;
-    }
-
-    /* pty masters are an exact copy of their slaves,
-    so no need to record them too. */
-    maj = K26_MAJOR(packet->p.dev);
-    return maj == 2   /* BSD ptm (/dev/ptypXY) */ ||
-           maj == 128 /* Unix98 ptm (-)        */;
-}
-
-static int packet_action(struct log_packet *packet, struct tty *tty, int fd) {
-    switch(packet->p.event) {
+static int packet_action(struct mem_packet *packet, struct tty *tty, int fd) {
+    switch(packet->event) {
        /* EV_OPEN is not handled since a logfile will automatically be opened
        in log_write() if necessary. */
        case EV_READ:
@@ -254,27 +292,104 @@ static int packet_action(struct log_packet *packet, struct tty *tty, int fd) {
 }
 
 //-----------------------------------------------------------------------------
+static const char *m_idx = "pqrstuvwxyzabcdef", *s_idx = "0123456789abcdef";
+
+static char *devname(uint32_t dev, char *buf, size_t count) {
+    /* This function returns a "simple name" which can be integrated into
+    rpld's log file name scheme. */
+    unsigned long minor = K26_MINOR(dev);
+
+    switch(K26_MAJOR(dev)) {
+        case 2: // BSD pty masters
+            snprintf(buf, count, "pty%c%c",
+             m_idx[minor >> 4], s_idx[minor % 16]);
+            break;
+        case 3: // BSD pty slaves
+            snprintf(buf, count, "tty%c%c",
+             m_idx[minor >> 4], s_idx[minor % 16]);
+            break;
+        case 4:
+            // Virtual console (0-63) and serial console (64-255)
+            if(minor < 64) { snprintf(buf, count, "vc-%lu", minor); }
+            else { snprintf(buf, count, "ttyS%lu", minor - 64); }
+            break;
+        case 5:
+            if(minor == 1) { snprintf(buf, count, "console"); }
+            else { goto unknown; }
+        case 128:
+            // Unix98 pty masters (got no real /dev entry, though, see below)
+            snprintf(buf, count, "ptm-%lu", minor);
+            break;
+        case 136: // Unix98 pty slaves
+            snprintf(buf, count, "pts-%lu", minor);
+            break;
+    }
+    return buf;
+
+unknown:
+    /* Not every char device is a tty, and I even do not want to include a
+    name database for all possible tty devices. */
+    snprintf(buf, count, "%lu-%lu", K26_MAJOR(dev), minor);
+    return buf;
+}
+
+static char *devname_fs(uint32_t dev, char *buf, size_t count) {
+    /* This function returns a device filename generated from major/minor
+    number, which is then used for stat(). */
+    unsigned long minor = K26_MINOR(dev);
+
+    switch(K26_MAJOR(dev)) {
+        case 2: // BSD pty masters
+            snprintf(buf, count, "/dev/pty%c%c",
+             m_idx[minor >> 4], s_idx[minor % 16]);
+            break;
+        case 3: // BSD pty slaves
+            snprintf(buf, count, "/dev/tty%c%c",
+             m_idx[minor >> 4], s_idx[minor % 16]);
+            break;
+        case 4:
+            // Virtual console (0-63) and serial console (64-255)
+            if(minor < 64) { snprintf(buf, count, "/dev/tty%lu", minor); }
+            else { snprintf(buf, count, "/dev/ttyS%lu", minor - 64); }
+            break;
+        case 5:
+            if(minor == 1) { snprintf(buf, count, "/dev/console"); }
+            else { goto unknown; }
+        case 128:
+            /* Unix98 pty masters. They usually have no device nodes in /dev,
+            and should instead be accessed over the multiplexer /dev/ptmx.
+            However IIRC, devfs popped them up in /dev/ptm. */
+            snprintf(buf, count, "/dev/ptm/%lu", minor);
+            break;
+        case 136: // Unix98 pty slaves
+            snprintf(buf, count, "/dev/pts/%lu", minor);
+            break;
+    }
+    return buf;
+
+unknown:
+    return NULL;
+}
+
 static struct tty *get_tty(uint32_t dev) {
     struct HXbtree_node *ts;
     struct tty *ret, *tty;
     if((ret = HXbtree_get(Ttys, (void *)dev)) != NULL) { return ret; }
     if((tty = malloc(sizeof(struct tty))) == NULL) { return NULL; }
     tty->dev = dev;
-    tty->in  = \
-    tty->out = 0;
     tty->fd  = -1;
     if((ts = HXbtree_add(Ttys, (void *)dev, tty)) == NULL) {
-        perror("Memory allocation failure");
+        notify(LOG_ERR, NF_NEWLINE, "Memory allocation failure");
         return NULL;
     }
     return ts->data;
 }
 
 static int log_open(struct tty *tty) {
-    char buf[MAXFNLEN], devname[64], fmday[16], fmtime[16], user[32];
+    char buf[MAXFNLEN], devnode[64], fmday[16], fmtime[16], user[32];
     char *subst_catalog[] = {
         "d", fmday,
-        "l", devname,
+        "l", devnode,
         "t", fmtime,
         "u", user,
         NULL
@@ -282,8 +397,8 @@ static int log_open(struct tty *tty) {
 
     { // User
       struct stat sb;
-      SH_devname_fs(tty->dev, devname, 64);
-      if(stat(devname, &sb) != 0) {
+      devname_fs(tty->dev, devnode, 64);
+      if(stat(devnode, &sb) != 0) {
           strcpy(user, "NONE");
       } else if(getnamefromuid(sb.st_uid, user, 32) == NULL) {
           snprintf(user, 32, "%d", sb.st_uid);
@@ -294,29 +409,54 @@ static int log_open(struct tty *tty) {
       time_t now = time(NULL);
       struct tm now_tm;
       localtime_r(&now, &now_tm);
-      strftime(fmday,  16, "%Y%M%d", &now_tm);
+      strftime(fmday,  16, "%Y%m%d", &now_tm);
       strftime(fmtime, 16, "%H%M%S", &now_tm);
     }
 
     // Simple tty name
-    SH_devname(tty->dev, devname, 64);
+    devname(tty->dev, devnode, 64);
     replace(buf, Opt.ofmt, MAXFNLEN, subst_catalog);
 
+    if(Opt.nocreat) { snprintf(buf, MAXFNLEN, "/dev/null"); }
     if((tty->fd = open(buf, O_WRONLY | O_CREAT | O_APPEND, 0600)) < 0) {
-        fprintf(stderr, "Could not open %s: %s\n", buf, strerror(errno));
+        notify(LOG_ERR, NF_NEWLINE, "Could not open %s: %s\n",
+         buf, strerror(errno));
     }
+
+    // ident header (optional)
+    if(tty->fd >= 0) {
+        struct disk_packet p = {
+            .event = EV_IDENT,
+            .magic = MAGIC_SIG,
+            .tv    = {-1, -1},
+        };
+        size_t s = snprintf(buf, MAXFNLEN, "ttyrpld %s", Opt._version);
+        p.size = s;
+#ifdef BIGENDIAN
+        SH_swab(&p.size, sizeof(uint16_t));
+#endif
+        write(tty->fd, &p, sizeof(struct disk_packet));
+        write(tty->fd, buf, s);
+    }
+
     return tty->fd;
 }
 
-static void log_write(struct log_packet *packet, struct tty *tty, int fd) {
+static void log_write(struct mem_packet *packet, struct tty *tty, int fd) {
+    size_t s = packet->size;
     int read_left;
 
     if(tty->fd == -1) { log_open(tty); }
-    write(tty->fd, packet, sizeof(struct log_packet));
-    read_left = packet->p.size;
+#ifdef BIGENDIAN
+    SH_swab(&packet->size, sizeof(packet->size));
+    SH_swab(&packet->tv.tv_sec, sizeof(packet->tv.tv_sec));
+    SH_swab(&packet->tv.tv_usec, sizeof(packet->tv.tv_usec));
+#endif
+    write(tty->fd, &packet->size, sizeof(struct disk_packet));
+    read_left = s;
 
     while(read_left > 0) {
-        int have_read = read(fd, Buffer, imin(read_left, Opt.bsize));
+        int have_read = read(fd, Buffer, min_uint(read_left, Opt.bsize));
         write(tty->fd, Buffer, have_read);
         read_left -= have_read;
     }
@@ -327,12 +467,13 @@ static void log_write(struct log_packet *packet, struct tty *tty, int fd) {
 static void log_close(struct tty *tty) {
     close(tty->fd);
     tty->fd = -1;
+    HXbtree_del(Ttys, tty);
     return;
 }
 
 //-----------------------------------------------------------------------------
 static int init_device(const char *in_devs) {
-    const char *new_dev;
+    const char *new_dev = NULL;
     int fd = -1;
 
     if(Opt.ovdev == NULL && (new_dev = mknod_rpldev()) != NULL) {
@@ -370,16 +511,8 @@ static int init_device(const char *in_devs) {
 static int init_sbuffer(int fd) {
     int eax;
 
-    if(Opt.bsize > 0) {
-        // nothing
-    } else if((eax = ioctl(fd, RPL_IOC_GETBUFSIZE, 0)) >= 0) {
-        // use ioctl() value if BSIZE is not given
+    if((eax = ioctl(fd, RPL_IOC_GETBUFSIZE, 0)) >= 0) {
         Opt.bsize = eax;
-    } else {
-        // use FBSIZE if ioctl() failed
-        fprintf(stderr, "Warning: Could not detect buffer size from RPL"
-         " device\n");
-        Opt.bsize = Opt.fbsize;
     }
 
     if(posix_memalign(&Buffer, sysconf(_SC_PAGESIZE), Opt.bsize) != 0) {
@@ -446,9 +579,13 @@ static int init_sighandler(void) {
     s_alrm.sa_handler = sighandler_alrm;
     s_alrm.sa_flags   = SA_RESTART;
     sigemptyset(&s_alrm.sa_mask);
-    
-    return sigaction(SIGINT, &s_int, NULL) + sigaction(SIGTERM, &s_int, NULL) +
-     sigaction(SIGALRM, &s_alrm, NULL);
+
+    /* All sigactions() shall be executed, however, if one fails, this function
+    shall return <= 0, otherwise >0 upon success.
+    Geesh, I love these constructs. */
+    return !(!!sigaction(SIGINT, &s_int, NULL) + 
+     !!sigaction(SIGTERM, &s_int, NULL) +
+     !!sigaction(SIGALRM, &s_alrm, NULL));
 }
 
 static void sighandler_int(int s) {
@@ -481,18 +618,18 @@ static int get_options(int *argc, const char ***argv) {
     static const char *_empty_argv[] = {NULL};
     char *tmp;
     struct poptOption options_table[] = {
-        {"bsize", 0, POPT_ARG_LONG, &Opt.bsize, 0,
-         "Buffer size (overrides ioctl()-retrieved size)", "bytes"},
-        {"fbsize", 0, POPT_ARG_LONG, &Opt.fbsize, 0,
-         "Buffer size in case ioctl() fails", "bytes"},
         {NULL, 'D', POPT_ARG_STRING, &Opt.ovdev, 0,
          "Path to the RPL device", "file"},
         {NULL, 'O', POPT_ARG_STRING, &Opt.ofmt, 0,
          "Override OFMT variable", "string"},
+        {NULL, 'Q', POPT_ARG_NONE, &Opt.nocreat, 0,
+         "Debug: do not create any files", NULL},
         {NULL, 'U', POPT_ARG_STRING, &tmp, 'U',
          "User to change to", "user"},
         {NULL, 'c', POPT_ARG_STRING, &tmp, 'c',
          "Read configuration from file (on top of hardcoded conf)", "file"},
+        {NULL, 's', POPT_ARG_NONE, &Opt.syslog, 0,
+         "Use syslog rather than stderr for messages", NULL},
         {NULL, 'v', POPT_ARG_NONE, &Opt.verbose, 0,
          "Print statistics while rpld is running", NULL},
         POPT_AUTOHELP
@@ -537,8 +674,6 @@ static int get_options(int *argc, const char ***argv) {
 
 static int read_config(const char *file) {
     static struct rconfig_opt options_table[] = {
-        {"BSIZE",    RCONF_LONG,   &Opt.bsize,  NULL},
-        {"FBSIZE",   RCONF_LONG,   &Opt.fbsize, NULL},
         {"OVDEVICE", RCONF_STRING, &Opt.ovdev,  NULL},
         {"FBDEVICE", RCONF_STRING, &Opt.fbdev,  NULL},
         {"OFMT",     RCONF_STRING, &Opt.ofmt,   NULL},
@@ -548,6 +683,19 @@ static int read_config(const char *file) {
     return HX_rconfig(file, options_table);
 }
 
+static int read_config_bp(const char *app_path, const char *file) {
+    char *fpath = HX_strdup(app_path), *ptr, construct[MAXFNLEN];
+    if((ptr = strrchr(fpath, '/')) == NULL) {
+        construct[MAXFNLEN - 1] = '\0';
+        strncpy(construct, file, MAXFNLEN - 1);
+    } else {
+        *ptr++ = '\0';
+        snprintf(construct, MAXFNLEN, "%s/%s", fpath, file);
+    }
+    free(fpath);
+    return read_config(construct);
+}
+
 static int pconfig_user(const char *key, unsigned char type, void *ptr,
  void *uptr) {
     if((Opt.user_id = getuidfromname(ptr)) < 0) {
@@ -576,8 +724,23 @@ static uid_t getuidfromname(const char *name) {
     return ep->pw_uid;
 }
 
-inline static int imin(int a, int b) {
-    return a < b ? a : b;
+static int notify(int lv, unsigned int opts, const char *fmt, ...) {
+    if(Opt.verbose) {
+        int rv;
+        va_list argp;
+        va_start(argp, fmt);
+        if(opts & NF_NEWLINE) { fprintf(stderr, "\n"); }
+        rv = vfprintf(stderr, fmt, argp);
+        va_end(argp);
+        return rv;
+    } else if(Opt.syslog) {
+        va_list argp;
+        va_start(argp, fmt);
+        vsyslog(lv, fmt, argp);
+        va_end(argp);
+        return 1;
+    }
+    return 0;
 }
 
 static char *replace(char *dest, const char *src, size_t count,
@@ -587,7 +750,7 @@ static char *replace(char *dest, const char *src, size_t count,
     destp[count - 1] = '\0';
 
     while(*src != '\0' && count > 1) {
-        size_t max = imin(strcspn(src, "%"), count - 1);
+        size_t max = min_uint(strcspn(src, "%"), count - 1);
         char **catp = catalog, *key, *value;
         int got_spec = 0;
 
@@ -604,7 +767,7 @@ static char *replace(char *dest, const char *src, size_t count,
             continue;
         }
         if((unsigned char)*src >= 128 || !isalpha(*src)) {
-            fprintf(stderr, "%% without specifier in OFMT.\n");
+            notify(LOG_WARNING, NF_NEWLINE, "%% without specifier in OFMT.\n");
             break;
         }
 
@@ -620,7 +783,9 @@ static char *replace(char *dest, const char *src, size_t count,
             }
         }
         if(count < 2) { break; }
-        if(!got_spec) { fprintf(stderr, "Invalid specifier: %%%c\n", *src); }
+        if(!got_spec) {
+            notify(LOG_WARNING, NF_NEWLINE, "Invalid specifier: %%%c\n", *src);
+        }
         ++src;
     }
     return dest;
